/* indent:4  tabsize:8  font:fixed-width */

#import "FreezeL.h"
#import "localstrings.h"
#import "FreezeLTempDelegate.h"
#import "FreezeLSuitDelegate.h"
#import "FreezeLGameDelegate.h"
#import "FreezeLPrefs.h"

static id _freezeLSharedInstance = nil;


@implementation FreezeL

+ (FreezeL*) sharedInstance
    /*"
    Instead of the old method of calling SolGameController when
     you wanted to message the FreezeL game module, you now
     call [FreezeL sharedInstance] and direct messages to
     the returned instance. You can use this to check if the game
     has been won, among other things.
     "*/
{
    return _freezeLSharedInstance;
}

- initFromBundle:(NSBundle*)aBundle withName:(NSString*)name
    /*"
    Extends our superclass method in order to set our shared
     instance variable.
     "*/
{
    [super initFromBundle:aBundle withName:name];

    // This is used instead of having a global SolGameController()
    // function. It is assumed that this method will be called
    // only once and that our +sharedInstance method will return
    // the single instance.
    _freezeLSharedInstance = self;

    return self;
}


/*---------------------------------------------------------------------------
|
|    - startGame:
|
|----------------------------------------------------------------------------
|
|    Start a new game.  Get confirmation from the user before aborting a
|    game in progress.
|			
\----------------------------------------------------------------------------*/

- (void) startGame:sender
{
    if (gameInProgress)
        [(FreezeLPrefs *)prefs accumulateLoss];
    
    [super startGame:sender];
    [self setupGame:YES];
}


/*---------------------------------------------------------------------------
|
|    - restartGame:
|
|----------------------------------------------------------------------------
|
|    Restart the game in progress.
|			
\----------------------------------------------------------------------------*/

- (void) restartGame:sender
{
    if (gameInProgress)
        [(FreezeLPrefs *)prefs accumulateLoss];
    
    [super restartGame:sender];
    [self setupGame:NO];
}


/*---------------------------------------------------------------------------
|
|    - setupGame:(BOOL)redeal
|
|----------------------------------------------------------------------------
|
|    Setup a new game.  If "redeal" is true, deal a new deck, otherwise
|    use the same cards as the previous game.
|			
\----------------------------------------------------------------------------*/

- (void) setupGame:(BOOL)redeal
{
    int 			pileIndex, i;
    int 			cycle;

    gameInProgress = YES;
    [gameWindow disableFlushWindow];

    //----------------------------------------------------------------
    //    Sync pile id's with current game window and set
    //    preferences and delegates
    //----------------------------------------------------------------

    suitCardPiles[0] = suitPileView1;
    suitCardPiles[1] = suitPileView2;
    suitCardPiles[2] = suitPileView3;
    suitCardPiles[3] = suitPileView4;
    for (i = 0; i < 4; i++)
      {
        [suitCardPiles[i] setBackgroundColor:desktopColor];
        [suitCardPiles[i] setCardSize:cardSize];
        [suitCardPiles[i] setDelegate:suitPileDelegate];
      }

    gameCardPiles[0] = gamePileView1;
    gameCardPiles[1] = gamePileView2;
    gameCardPiles[2] = gamePileView3;
    gameCardPiles[3] = gamePileView4;
    gameCardPiles[4] = gamePileView5;
    gameCardPiles[5] = gamePileView6;
    gameCardPiles[6] = gamePileView7;
    gameCardPiles[7] = gamePileView8;
    for (i = 0; i < 8; i++)
      {
        [gameCardPiles[i] setBackgroundColor:desktopColor];
        [gameCardPiles[i] setCardSize:cardSize];
        [gameCardPiles[i] setDelegate:gamePileDelegate];
      }

    tempCardPiles[0] = tempPileView1;
    tempCardPiles[1] = tempPileView2;
    tempCardPiles[2] = tempPileView3;
    tempCardPiles[3] = tempPileView4;
    for (i = 0; i < 4; i++)
      {
        [tempCardPiles[i] setBackgroundColor:desktopColor];
        [tempCardPiles[i] setCardSize:cardSize];
        [tempCardPiles[i] setDelegate:tempPileDelegate];
      }

    [gamePileDelegate setSuitCardPileViews:suitCardPiles];
    [gamePileDelegate setTempCardPileViews:tempCardPiles];
    [tempPileDelegate setSuitCardPileViews:suitCardPiles];

    [gameWindow display];
    [gameWindow enableFlushWindow];
    [gameWindow flushWindow];

    //-----------------------------------------------------------------------
    //    Initialize the gamePile to have a 1 shuffled deck.
    //-----------------------------------------------------------------------

    // Initialize the initial cardPile.
    if (!drawCardPile)
      {
        drawCardPile = [[CardPile allocWithZone:[self zone]]
                    initForCardSize:cardSize];
      }
    else
        [drawCardPile setCardSize:cardSize];

    [drawCardPile empty];
    if (redeal)
      {
        [drawCardPile addDeck];
        [drawCardPile shuffle];

        // make a copy of the CardPile for restart option
        if (!prevDeck)
          {
            prevDeck = [[CardPile allocWithZone:[self zone]]
                               initForCardSize:cardSize];
          }
        else
          {
            [prevDeck empty];
            [prevDeck setCardSize:cardSize];
          }
        [prevDeck addCopyOfPile:drawCardPile];
      }
    else
      {
        if (prevDeck)
          {
            // copy the saved deck back to the game deck
            [prevDeck setCardSize:cardSize];
            [drawCardPile addCopyOfPile:prevDeck];
          }
        else
          {
            // this shouldn't happen, but just in case...
            [drawCardPile empty];
            [drawCardPile addDeck];
            [drawCardPile addDeck];
            [drawCardPile shuffle];
          }
      }

    //-----------------------------------------------------------------------
    //    Initialize the "suit piles" as empty
    //-----------------------------------------------------------------------

    for (pileIndex = 0; pileIndex < 4; pileIndex++)
      {
        [[suitCardPiles[pileIndex] cardPile] empty];
        [suitCardPiles[pileIndex] display];
      }

    //-----------------------------------------------------------------------
    //    Initialize the "temp piles" as empty
    //-----------------------------------------------------------------------

    for (pileIndex = 0; pileIndex < 4; pileIndex++)
      {
        [[tempCardPiles[pileIndex] cardPile] empty];
        [tempCardPiles[pileIndex] display];
      }

    //-----------------------------------------------------------------------
    //    Initialize and deal cards to the 8 "user piles"
    //-----------------------------------------------------------------------

    for (pileIndex = 0; pileIndex < 8; pileIndex++)
      {
        [[gameCardPiles[pileIndex] cardPile] empty];
        [gameCardPiles[pileIndex] display];
      }

    for (cycle = 0; cycle < 6; cycle++)
      {
        for (pileIndex = 0; pileIndex < 8; pileIndex++)
          {
            CardPile*	 userPile = [gameCardPiles[pileIndex] cardPile];
            Card* 		tempCard = [drawCardPile dealTopCard];
            [userPile insertCard:tempCard at:CS_TOP];
            [[userPile topCard] flip];
            [gameCardPiles[pileIndex] display];
          }
      }
    
    for (pileIndex = 0; pileIndex < 4; pileIndex++)
      {
        CardPile* 	userPile = [gameCardPiles[pileIndex] cardPile];
        Card* 		tempCard = [drawCardPile dealTopCard];
        [userPile insertCard:tempCard at:CS_TOP];
        [[userPile topCard] flip];
        [gameCardPiles[pileIndex] display];
      }

    [gameWindow display];
    [gameWindow makeKeyAndOrderFront:self];
    [self moveSuitableCards];
}

- (BOOL)searchForCardInCardPileViewArray:(CardPileView **)aCardPileViewArray
                                  length:(int)len {
    int i, j, k;

    for (i = 0; i < len; i++) {
        Card *t = [[aCardPileViewArray[i] cardPile] topCard];
        if (t) {
            for (j = 0; j < 4; j++) {
                Card *s = [[suitCardPiles[j] cardPile] topCard];
                if ((s == nil && [t value] == CS_ACE)
                    || (s != nil && [t suit] == [s suit] && [t value] == [s value] + 1)) {
                    // it's legal to move t to this suit pile, now see if it's desirable
                    if ([t value] > CS_TWO) {
                        int points = 0;
                        
                        for (k = 0; k < 4; k++) {
                            Card *p = [[suitCardPiles[k] cardPile] topCard];
                            if (p != nil && ([p cardColor] != [t cardColor])
                                && ((int)([t value] - [p value]) <= 2)) {
                                points++;
                            }
                        }
                        
                        if (points < 2)
                            break;
                    }
                    
                    // do the move
                    [[suitCardPiles[j] cardPile] addCard:t];
                    [suitCardPiles[j] display];
                    [[aCardPileViewArray[i] cardPile] removeCard:t];
                    [aCardPileViewArray[i] display];
                    
                    return YES;
                }
            }
        }
    }
    return NO;
}

- (void)moveSuitableCards
{
    while ([self searchForCardInCardPileViewArray:gameCardPiles length:8]
           || [self searchForCardInCardPileViewArray:tempCardPiles length:4]) {
        ; /* DO NOTHING */
    }

    [self checkForWin];
}

/*---------------------------------------------------------------------------
|
|    - endGame:sender
|
|----------------------------------------------------------------------------
|
|    End the game in progress.  Discard the game window.
|			
\----------------------------------------------------------------------------*/

- (void) endGame:sender
{
    if (gameInProgress)
        [(FreezeLPrefs *)prefs accumulateLoss];
    
    // close the game window
    [super endGame:sender];

    // ****custom code here****
}

/*---------------------------------------------------------------------------
|
|    - win
|
|----------------------------------------------------------------------------
|
|    Called when the game has been won.  This is where you can insert fancy
|    winning routines, or just call the default (boring) routine.
|
\----------------------------------------------------------------------------*/

- (void) win
{
    gameInProgress = NO;
    [(FreezeLPrefs *)prefs accumulateWin];
    [super win];  // replace this with something wonderful
}


/*---------------------------------------------------------------------------
|
|    - checkForWin
|
|----------------------------------------------------------------------------
|
|    Called to check the state of the game.  Always override (unless your
                                                              |    game is impossible to win).
|
\----------------------------------------------------------------------------*/

- (void) checkForWin
{
    if ([[suitPileView1 cardPile] cardCount] +
        [[suitPileView2 cardPile] cardCount] +
        [[suitPileView3 cardPile] cardCount] +
        [[suitPileView4 cardPile] cardCount] == 52)
      {
        [self win];
      }
}

@end

